//---------------------------------------------------------------------------
//
//   MorphChecker
//
//          Copyright(C) 2014, Tark
//
//  [t`FbJ[
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include "MQPlugin.h"
#if MQPLUGIN_VERSION >= 0x0400
#include "MQWidget.h"
#endif
#include "MQSetting.h"

#include <algorithm>

#include "MorphCheckerWindow.h"
#include "MorphChecker.h"

static MorphChecker plugin;

//---------------------------------------------------------------------------
//  MQGetPlugInID
//---------------------------------------------------------------------------
void MorphChecker::GetPlugInID(DWORD *Product, DWORD *ID)
{
	*Product = 0xE3F7A830;
	*ID      = 0xBD9A1876;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//---------------------------------------------------------------------------
const char *MorphChecker::GetPlugInName(void)
{
	return "MorphChecker ver.1.0 Copyright(C) 2014, Tark.";
}

//---------------------------------------------------------------------------
//  MQEnumString
//---------------------------------------------------------------------------
const char *MorphChecker::EnumString()
{
	return "MorphChecker";
}

//---------------------------------------------------------------------------
//  Initialize
//---------------------------------------------------------------------------
BOOL MorphChecker::Initialize()
{
	return TRUE;
}

//---------------------------------------------------------------------------
//  Exit
//---------------------------------------------------------------------------
void MorphChecker::Exit()
{
	if(m_Window) {
		posX = m_Window->GetPosX();
		posY = m_Window->GetPosY();
		height = m_Window->GetHeight();
		width = m_Window->GetWidth();
		saveSetting();
	}
	if (m_Window){
		delete m_Window;
		m_Window = NULL;
	}
}

void MorphChecker::loadSetting()
{
	MQSetting *setting = OpenSetting();
	setting->Load("PosX", posX, -1);
	setting->Load("PosY", posY, -1);
	setting->Load("Height", height, -1);
	setting->Load("Width", width, -1);
	CloseSetting(setting);
}

void MorphChecker::saveSetting()
{
	MQSetting *setting = OpenSetting();
	setting->Save("PosX", posX);
	setting->Save("PosY", posY);
	setting->Save("Height", height);
	setting->Save("Width", width);
	CloseSetting(setting);
}

//---------------------------------------------------------------------------
//  Activate
//---------------------------------------------------------------------------
BOOL MorphChecker::Activate(MQDocument doc_, BOOL flag)
{
	if (!m_Window) {
		loadSetting();
		MQWindowBase mainwnd = MQWindow::GetMainWindow();
		m_Window = new MorphCheckerWindow(this, mainwnd);
		if(posX >= 0 && posY >= 0) {
			m_Window->SetPosX(posX);
			m_Window->SetPosY(posY);
			if(height > 0 && width > 0) {
				m_Window->SetHeight(height);
				m_Window->SetWidth(width);
			}
		}
	}
	m_bActivate = flag ? true : false;
	m_Window->SetVisible(m_bActivate);
	if(m_bActivate) {
		// SIuWFNg̉tOAbNtOۑAAbN
		m_objvisible.clear();
		m_objlock.clear();
		cmq::CMQDocument doc(doc_);
		int ocount = doc.getObjectCount();
		m_objvisible.reserve(ocount);
		m_objlock.reserve(ocount);
		for(cmq::oindex_t oindex = 0; oindex < ocount; oindex++) {
			cmq::CMQObject obj(doc.getObject(oindex));
			if(!obj.isLive()) {
				continue;
			}
			m_objvisible.insert(std::make_pair(oindex, obj.getVisible()));
			m_objlock.insert(std::make_pair(oindex, obj.getLocking()));
			obj.setVisible(false);
			obj.setLocking(true);
		}
		make_objectlist(doc_);
	}
	else {
		m_objupdate = false;
		m_morphupdate = false;
		delete_viewobject(doc_);
		// SIuWFNg̉tOAbNtO𕜌
		cmq::CMQDocument doc(doc_);
		int ocount = doc.getObjectCount();
		for(cmq::oindex_t oindex = 0; oindex < ocount; oindex++) {
			cmq::CMQObject obj(doc.getObject(oindex));
			if(!obj.isLive()) {
				continue;
			}
			auto findit = m_objvisible.find(oindex);
			if(findit != m_objvisible.end()) {
				obj.setVisible(findit->second);
			}
			findit = m_objlock.find(oindex);
			if(findit != m_objlock.end()) {
				obj.setLocking(findit->second);
			}
		}
		m_objvisible.clear();
		m_objlock.clear();
	}
	UpdateUndo();
	MQ_RefreshView(nullptr);
	return m_bActivate;
}

//---------------------------------------------------------------------------
//  IsActivated
//---------------------------------------------------------------------------
BOOL MorphChecker::IsActivated(MQDocument doc)
{
	if (m_Window == NULL) {
		return FALSE;
	}
	return m_bActivate;
}

// vOC̃x[XNXԂ
MQBasePlugin *GetPluginClass()
{
	return &plugin;
}

MorphChecker::MorphChecker() :
	m_Window(nullptr), m_bActivate(false), m_doc(nullptr),
	m_objupdate(false), m_morphupdate(false), m_editmode(false), m_visibility(DRAW_OBJECT_FACE),
	m_select_base(nullptr), m_select_morph(nullptr), m_select_vindex(-1), m_select_vindex_symmetry(-1)
{
}

MorphChecker::~MorphChecker()
{
	delete_viewobject(m_doc);
	if(m_Window){
		delete m_Window;
		m_Window = nullptr;
	}
}

// IuWFNgXgXV
void MorphChecker::make_objectlist(MQDocument doc_)
{
	if(!m_bActivate) {
		return;
	}
#if _DEBUG
	OutputDebugString(L"make_objectlist start\n");
#endif
	cmq::CMQDocument doc(doc_);
	std::vector<CMQBaseObject> baseobjlist;
	std::vector<CMQMorphObject> morphobjlist;
	// x[XIuWFNgƃ[tIuWFNg𒊏oăXg쐬
	int ocount = doc.getObjectCount();
	for(cmq::oindex_t oindex = 0; oindex < ocount; oindex++) {
		cmq::CMQObject obj(doc.getObject(oindex));
		if(!obj.isLive()) {
			continue;
		}
		std::string name = obj.getName();
		if(name[0] == '#' || name.substr(0, 4) == "bone" || name.substr(0, 6) == "anchor") {
			// ΏۊO
			continue;
		}
		if(name.substr(0, 5) == "elem:") {
			// [tIuWFNg
			CMQMorphObject morphobj(obj);
			auto findit = std::find_if(morphobjlist.begin(), morphobjlist.end(), [&](CMQMorphObject &mobj){return morphobj.getUniqueID() == mobj.getUniqueID();});
			if(findit == morphobjlist.end()) {
				auto findit2 = std::find_if(m_morphobjlist.begin(), m_morphobjlist.end(), [&morphobj](CMQMorphObject &old) {
					return morphobj.getName() == old.getName();
				});
				// Ỏʂł̃[tlp
				if(findit2 != m_morphobjlist.end()) {
					morphobj.setvalue(findit2->value());
				}
				morphobjlist.push_back(morphobj);
			}
		}
		else {
			// x[XIuWFNg
			CMQBaseObject baseobj(obj);
			auto findit = std::find_if(baseobjlist.begin(), baseobjlist.end(), [&](CMQBaseObject &bobj){return baseobj.getUniqueID() == bobj.getUniqueID();});
			if(findit == baseobjlist.end()) {
				auto findit2 = std::find_if(m_baseobjlist.begin(), m_baseobjlist.end(), [&baseobj](CMQBaseObject &old) {
					return baseobj.getName() == old.getName();
				});
				// Ỏʂł̕\tOp
				if(findit2 != m_baseobjlist.end()) {
					baseobj.setvisible(findit2->visible());
				}
				baseobjlist.push_back(baseobj);
			}
		}
	}
	m_objupdate = true;
	m_baseobjlist.swap(baseobjlist);
	m_morphobjlist.swap(morphobjlist);
	// [tIuWFNgƃx[XIuWFNg̃Nt
	for(auto &morphobj : m_morphobjlist) {
		for(auto &baseobj : m_baseobjlist) {
			if(morphobj.getObjName() == baseobj.getObjName()) {
				morphobj.setBase(baseobj.getUniqueID());
				baseobj.hasMorph(true);
			}
		}
	}
	m_Window->morphlistup();
	m_Window->listupdate(doc_);
#if _DEBUG
	OutputDebugString(L"make_objectlist end\n");
#endif
}

void MorphChecker::createDrawObject(MQDocument doc_, MQScene scene)
{
	if(!m_objupdate) {
		return;
	}
#if _DEBUG
	OutputDebugString(L"createDrawObject start\n");
#endif
	HCURSOR hcur = (HCURSOR)LoadImage(NULL, IDC_WAIT, IMAGE_CURSOR, 0, 0, LR_SHARED);
	SetCursor(hcur);

	cmq::CMQDocument doc(doc_);
	delete_viewobject(doc_);
	m_viewobjlist.reserve(m_baseobjlist.size());
	// x[XIuWFNgXg̃IuWFNgRs[ĕ`pIuWFNgɓo^
	for(auto &baseobj : m_baseobjlist) {
#if _DEBUG
		OutputDebugString(L".");
#endif
		SCENE_OPTION option;
		GetSceneOption(scene, option);
		DWORD  visibility = 0;
		if(option.ShowVertex) {visibility |= DRAW_OBJECT_POINT;}
		if(option.ShowLine) {visibility |= DRAW_OBJECT_LINE;}
		if(option.ShowFace) {visibility |= DRAW_OBJECT_FACE;}
		m_visibility = visibility;
		MQObject drawobj_ = CreateDrawingObject(doc_, static_cast<DRAW_OBJECT_VIISIBILITY>(m_visibility), false);
		cmq::CMQObject drawobj(drawobj_, doc);
		// x[XIuWFNg̃Rs[
		drawobj.copyFromObject(baseobj);
		// viewobjlistɓo^
		CMQViewObject viewobj(drawobj);
		m_viewobjlist.insert(std::make_pair(baseobj.getUniqueID(), viewobj));
	}
	m_objupdate = false;
	m_morphupdate = true;
	m_select_vindex = -1;
#if _DEBUG
	OutputDebugString(L"createDrawObject end\n");
#endif
	hcur = (HCURSOR)LoadImage(NULL, IDC_ARROW, IMAGE_CURSOR, 0, 0, LR_SHARED);
	SetCursor(hcur);
}

void MorphChecker::OnDraw(MQDocument doc_, MQScene scene, int width, int height)
{
	if(!m_bActivate) {
		return;
	}

	SCENE_OPTION option;
	GetSceneOption(scene, option);
	DWORD  visibility = 0;
	if(option.ShowVertex) {visibility |= DRAW_OBJECT_POINT;}
	if(option.ShowLine) {visibility |= DRAW_OBJECT_LINE;}
	if(option.ShowFace) {visibility |= DRAW_OBJECT_FACE;}
	if(m_visibility != visibility) {
		m_visibility = visibility;
		m_morphupdate = true;
	}

	// drawobjectXV
	createDrawObject(doc_, scene);
	// morphingXV
	morph(scene);

	if(!m_select_base || !m_select_morph) {
		return;
	}

	if(m_select_morph && m_select_morph->isLive()) {
		cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(m_select_morph);
		for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
			cmq::vindex_t vindex = vit.index();
			CMQMorphObject &mobj(*m_select_morph);
			bool sel = mobj.isSelectVertex(vindex);
			UINT uid = mobj.getUniqueID();
			UINT buid = mobj.getBase();
		}
	}

	// I𒸓_QƗpviewobj
	// ̃IuWFNg̒_̈ʒuőI𒸓_`悷
	auto findit = m_viewobjlist.find(m_select_base->getUniqueID());
	if(findit == m_viewobjlist.end()) {
		return;
	}
	CMQViewObject &viewobj = findit->second;

	// morphobj̑Iςݒ_vewobjɕ\
	cmq::CMQDocument doc(doc_);
	if(m_select_morph && m_select_morph->isLive()) {
		cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(m_select_morph);
		MQObject _draw = nullptr;
		cmq::CMQObject draw;
		for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
			if(m_select_morph->isSelectVertex(vit.index())) {
				if(!_draw) {
					_draw = CreateDrawingObject(doc_, DRAW_OBJECT_POINT);
					draw = cmq::CMQObject(_draw, nullptr);
					_draw->SetColor(MQColor(0,1,1));
					_draw->SetColorValid(TRUE);
				}
				MQPoint p1(viewobj.getVertex(vit.index()));
				draw.createLine(p1, p1);
			}
		}
	}
	// HIT_\
	if(m_select_vindex >= 0){
		MQObject _draw = CreateDrawingObject(doc_, DRAW_OBJECT_POINT);
		cmq::CMQObject draw(_draw, nullptr);
		_draw->SetColor(MQColor(1,1,0));
		_draw->SetColorValid(TRUE);
		MQPoint p1(viewobj.getVertex(m_select_vindex));
		draw.createLine(p1, p1);
	}
}

void MorphChecker::delete_viewobject(MQDocument doc)
{
	for(auto &viewpair : m_viewobjlist) {
		DeleteDrawingObject(doc, viewpair.second.m_obj);
	}
	m_viewobjlist.clear();
	m_select_viewobjlist.clear();
	m_select_base = nullptr;
	m_select_morph = nullptr;
	m_select_vindex = -1;
}

void MorphChecker::OnNewDocument(MQDocument doc, const char *filename, NEW_DOCUMENT_PARAM& param)
{
	m_doc = doc;
}

void MorphChecker::OnEndDocument(MQDocument doc)
{
	m_objvisible.clear();
	m_objlock.clear();
	m_baseobjlist.clear();
	m_morphobjlist.clear();
	delete_viewobject(doc);
}

void MorphChecker::OnUpdateObjectList(MQDocument doc)
{
	if(m_bActivate) {
		make_objectlist(doc);
	}
}

BOOL MorphChecker::OnUndo(MQDocument doc, int undo_state)
{
	if(m_bActivate) {
		make_objectlist(doc);
	}
	return FALSE;
}

BOOL MorphChecker::OnRedo(MQDocument doc, int undo_state)
{
	if(m_bActivate) {
		make_objectlist(doc);
	}
	return FALSE;
}

BOOL MorphChecker::OnMouseMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if(!m_bActivate) {
		return FALSE;
	}

	// ҏWIvV擾
	EDIT_OPTION option;
	GetEditOption(option);
	// Ώە̌mđI
	HIT_TEST_PARAM param;
	param.TestVertex = TRUE;
	param.TestLine = param.TestFace = FALSE;
	if(HitTestObjects(scene, state.MousePos, m_select_viewobjlist, param)) {
		if(param.HitType == HIT_TYPE_VERTEX){
			m_select_vindex = param.VertexIndex;
		}
		else {
			m_select_vindex = -1;
		}
	}
	RedrawScene(scene);
	return TRUE;
}

BOOL MorphChecker::OnLeftButtonDown(MQDocument doc_, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if(!m_bActivate || !m_editmode || !state.LButton) {
		return FALSE;
	}
	m_select_vindex = -1;

	std::vector<CMQMorphObject*> morphobjsublist;
	if(m_select_morph) {
		// vZΏۃ[tIuWFNg𒊏o
		morphobjsublist.reserve(m_morphobjlist.size());
		for(auto &morphobj : m_morphobjlist) {
			if(!morphobj.isLive()) {
				continue;
			}
			// wl0łȂ[t邩
			if(morphobj.getBase() == m_select_morph->getBase()) {
				morphobjsublist.push_back(&morphobj);
			}
		}
	}

	// ҏWIvV擾
	EDIT_OPTION option;
	GetEditOption(option);

	// Ώە̌mđI
	HIT_TEST_PARAM param;
	param.TestVertex = TRUE;
	param.TestLine = param.TestFace = FALSE;
	if(HitTestObjects(scene, state.MousePos, m_select_viewobjlist, param)) {
		if(param.HitType == HIT_TYPE_VERTEX){
			// Hitviewobj̒_ɑΉmorphobj̒_I
			if(m_select_morph && m_select_morph->getUniqueID() == m_Window->getCurrentMorph()) {
				if(!state.Shift) {
					// ShiftĂȂǉIł͂Ȃ̂ŃNA
					doc_->ClearSelect(MQDOC_CLEARSELECT_ALL);
				}
				doc_->AddSelectVertex(doc_->GetObjectIndex(m_select_morph->m_obj), param.VertexIndex);
				m_select_vindex = param.VertexIndex;
				MQPoint bpoint = m_select_base->getVertex(m_select_vindex);
				MQPoint vpoint0 = bpoint;
				m_select_mpoint0 = morph_one(vpoint0, bpoint, m_select_vindex, morphobjsublist, m_select_morph, false);
				m_select_mpoint100 = morph_one(vpoint0, bpoint, m_select_vindex, morphobjsublist, m_select_morph, true);
				if(option.Symmetry) {
					// Ώ̒_
					cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(m_select_morph);
					MQPoint vs(m_select_morph->getVertex(m_select_vindex));
					MQMatrix r;
					if(option.CoordinateType == COORDINATE_LOCAL) {
						r = m_select_morph->getLocalInverseMatrix();
						vs = vs * r;
					}
					vs.x = -vs.x;
					float dist = std::powf(0.5f, 2); // option.SymmetryDistance
					m_select_vindex_symmetry = -1;
					for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
						MQPoint v = m_select_morph->getVertex(vit.index());
						if(option.CoordinateType == COORDINATE_LOCAL) {
							v = v * r;
						}
						float d = (v - vs).norm();
						if(d < dist) {
							dist = d;
							m_select_vindex_symmetry = vit.index();
						}
					}
				}
#if _DEBUG
	OutputDebugString(L"OnLeftButtonDown: UpdateUndo\n");
#endif
				UpdateUndo();
				return TRUE;
			}
		}
	}
	return FALSE;
}

BOOL MorphChecker::OnLeftButtonMove(MQDocument doc_, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if(!m_bActivate || !m_editmode || !state.LButton) {
		return FALSE;
	}
	if(m_select_vindex == -1 || !m_select_morph || m_select_morph->getUniqueID() != m_Window->getCurrentMorph() || m_select_morph->value() == 0.0) {
		return FALSE;
	}

	// ҏWIvV擾
	EDIT_OPTION option;
	GetEditOption(option);

	auto findit = m_viewobjlist.find(m_select_base->getUniqueID());
	if(findit == m_viewobjlist.end()) {
		return FALSE;
	}
	auto &viewobj = findit->second;
	float out_w = 0;
	MQMatrix r, r2;
	if(option.CoordinateType == COORDINATE_LOCAL) {
		r = m_select_morph->getLocalInverseMatrix();
		r2 = m_select_morph->getLocal();
	}
	// Y_̃XN[WZێ܂܃}EX̍Wł3DW߂
	MQPoint vpos1 = viewobj.getVertex(m_select_vindex);
	MQPoint spos1 = scene->Convert3DToScreen(vpos1, &out_w);
	MQPoint spos2((float)state.MousePos.x, (float)state.MousePos.y, spos1.z);
	MQPoint vpos2 = scene->ConvertScreenTo3D(spos2);
	if(option.CoordinateType == COORDINATE_LOCAL) {
		vpos1 = vpos1 * r;
		vpos2 = vpos2 * r;
	}
	if(!option.SnapX) {
		vpos2.x = vpos1.x;
	}
	if(!option.SnapY) {
		vpos2.y = vpos1.y;
	}
	if(!option.SnapZ) {
		vpos2.z = vpos1.z;
	}
	if (option.SnapGrid){
		// Obhւ̋z
		vpos2 = GetSnappedPos(scene, vpos2, option.SnapGrid);
	}
	MQPoint vposvec = (vpos2 - vpos1) * (float)(100.0 / m_select_morph->value());
	MQPoint mpos = m_select_morph->getVertex(m_select_vindex);
	if(option.CoordinateType == COORDINATE_LOCAL) {
		mpos = mpos * r;
	}
	mpos += vposvec;
	if(option.Symmetry && m_select_vindex_symmetry != -1) {
		// Ώ̕ҏW
		MQPoint possym(mpos);
		possym.x = -possym.x;
		if(option.CoordinateType == COORDINATE_LOCAL) {
			possym = possym * r2;
		}
		m_select_morph->setVertex(m_select_vindex_symmetry, possym);
	}
	if(option.CoordinateType == COORDINATE_LOCAL) {
		mpos = mpos * r2;
	}
	m_select_morph->setVertex(m_select_vindex, mpos);

	m_morphupdate = true;
	RedrawScene(scene);
	return TRUE;
}

void MorphChecker::morph(MQScene scene)
{
	if(!m_morphupdate) {
		return;
	}
	// [tKp
	for(auto &baseobj : m_baseobjlist) {
		if(!baseobj.isLive()) {
			continue;
		}
		auto findit = m_viewobjlist.find(baseobj.getUniqueID());
		if(findit == m_viewobjlist.end()) {
			continue;
		}
		// Ԑݒi_AӁAʕ\tOfj
		CMQViewObject &viewobj(findit->second);
		viewobj.setVisible(baseobj.visible() ? m_visibility : 0);

		if(!baseobj.hasMorph() || !baseobj.visible()) {
			continue;
		}
		// vZΏۃ[tIuWFNg𒊏o
		std::vector<CMQMorphObject*> morphobjsublist;
		morphobjsublist.reserve(m_morphobjlist.size());
		for(auto &morphobj : m_morphobjlist) {
			if(!morphobj.isLive()) {
				continue;
			}
			// wl0łȂ[t邩
			if(morphobj.ismorphed() && baseobj.getUniqueID() == morphobj.getBase()) {
				morphobjsublist.push_back(&morphobj);
			}
		}
		if(morphobjsublist.empty()) {
			// [tύXĂȂ̂ŌvZsv
			continue;
		}

		cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(&baseobj);
		for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
			// baseobj̒_ʒuviewobj̏
			MQPoint bpoint = baseobj.getVertex(vit.index());
			MQPoint vpoint0 = bpoint;
			// wIuWFNgɑ΂郂[t̕ωʂZ
			vpoint0 = morph_one(vpoint0, bpoint, vit.index(), morphobjsublist);
			// ωʂviewobjɓKp
			viewobj.setVertex(vit.index(), vpoint0);
			viewobj.morphed();
		}
	}
	m_morphupdate = false;
}

MQPoint MorphChecker::morph_one(MQPoint &vpoint0, MQPoint &bpoint, cmq::vindex_t vindex, std::vector<CMQMorphObject*> &morphobjsublist, CMQMorphObject *except, bool full)
{
	for(auto *pmorphobj : morphobjsublist) {
		if(!full && pmorphobj == except) {
			continue;
		}
		// [tl0łȂ̂ɂĉZ
		MQPoint mpoint = pmorphobj->getVertex(vindex);
		if(mpoint != bpoint) {
			MQPoint diff = (full && pmorphobj == except) ? (mpoint - bpoint) : (mpoint - bpoint) * (float)(pmorphobj->value() / 100.0);
			vpoint0 += diff;
		}
	}
	return vpoint0;
}

void MorphChecker::set_morph(UINT uid, double value)
{
	for(auto &morphobj : m_morphobjlist) {
		if(morphobj.isLive() && morphobj.getUniqueID() == uid) {
			morphobj.setvalue(value);
			m_morphupdate = true;
			break;
		}
	}
}

double MorphChecker::get_morph_value(UINT uid)
{
	for(auto &morphobj : m_morphobjlist) {
		if(morphobj.isLive() && morphobj.getUniqueID() == uid) {
			return morphobj.value();
		}
	}
	return 0.0;
}

void MorphChecker::set_visible(UINT uid, bool value)
{
	for(auto &baseobj : m_baseobjlist) {
		if(baseobj.isLive() && baseobj.getUniqueID() == uid) {
			baseobj.setvisible(value);
			auto findit = m_viewobjlist.find(baseobj.getUniqueID());
			if(findit != m_viewobjlist.end()) {
				CMQViewObject &viewobj(findit->second);
				viewobj.setVisible(baseobj.visible() ? m_visibility : 0);
//				MQ_RefreshView(nullptr);
			}
			break;
		}
	}
}

bool MorphChecker::get_visible(UINT uid)
{
	for(auto &baseobj : m_baseobjlist) {
		if(baseobj.isLive() && baseobj.getUniqueID() == uid) {
			return baseobj.visible();
		}
	}
	return false;
}

UINT MorphChecker::getBaseUIDfromMorphUID(MQDocument doc_, UINT morphuid)
{
	MQObject mobj_ = doc_->GetObjectFromUniqueID(morphuid);
	auto match = std::find_if(m_morphobjlist.begin(), m_morphobjlist.end(), [&mobj_](const CMQMorphObject &mobj) {return mobj.m_obj == mobj_;});
	if(match != m_morphobjlist.end()) {
		UINT baseuid = match->getBase();
		return baseuid;
	}
	return -1;
}

void MorphChecker::setSelectingMorph(UINT uid)
{
	m_select_morph = nullptr;
	m_select_base = nullptr;
	m_select_viewobjlist.clear();
	if(m_editmode) {
		{
			auto findit = std::find_if(m_morphobjlist.begin(), m_morphobjlist.end(), [&uid](CMQMorphObject &mobj) {return mobj.getUniqueID() == uid;});
			m_select_morph = (findit != m_morphobjlist.end()) ? &(*findit) : nullptr;
		}
		if(m_select_morph) {
			UINT buid = m_select_morph->getBase();
			auto findit = std::find_if(m_baseobjlist.begin(), m_baseobjlist.end(), [&buid](CMQBaseObject &bobj) {return bobj.getUniqueID() == buid;});
			m_select_base = findit != m_baseobjlist.end() ? &(*findit) : nullptr;
		}
		if(m_select_base) {
			auto findit = m_viewobjlist.find(m_select_base->getUniqueID());
			if(findit != m_viewobjlist.end()) {
				m_select_viewobjlist.push_back(findit->second.m_obj);
			}
		}
	}
}

//--------------------------------------------------------------------------------

CMQBaseObject::CMQBaseObject(const cmq::CMQObject &obj) : cmq::CMQObject(obj), m_hasmorph(false), m_visible(true)
{
	std::wstring name = cmq::towstring(getName());
	std::wstring::size_type pos;
	if((pos = name.find(':')) != std::string::npos) {
		m_objname = name.substr(pos + 1);
	}
	if((pos = name.find('-')) != std::string::npos && pos > 1) {
		m_objname = name.substr(0, pos);
	}
	if(m_objname.empty()) {
		m_objname = name;
	}
}

CMQBaseObject &CMQBaseObject::operator=(const CMQBaseObject &other)
{
	if(this == &other) {return *this;}
	CMQObject::operator=(other);
	this->m_objname = other.m_objname;
	return *this;
}

CMQMorphObject::CMQMorphObject(const cmq::CMQObject &obj) : cmq::CMQObject(obj), m_value(0.0), m_baseuid(-1), m_morphed(false)
{
	std::wstring name = cmq::towstring(getName());
	std::wstring::size_type pos;
	if((pos = name.find('-')) != std::string::npos) {
		m_objname = name.substr(pos + 1);
		if(name.substr(0, 5) == L"elem:") {
			m_morphname = name.substr(5, pos - 5);
		}
	}
}

CMQMorphObject &CMQMorphObject::operator=(const CMQMorphObject &other)
{
	if(this == &other) {return *this;}
	CMQObject::operator=(other);
	this->m_morphname = other.m_morphname;
	this->m_objname = other.m_objname;
	this->m_value = other.m_value;
	this->m_baseuid = other.m_baseuid;
	this->m_morphed = other.m_morphed;
	return *this;
}
