#include <windows.h>
#include <algorithm>

#include "MorphCheckerWindow.h"
#include "MorphChecker.h"

MorphCheckerWindow::MorphCheckerWindow(MorphChecker *plugin, MQWindowBase& parent) : MQWindow(parent), m_pPlugin(plugin),
		m_current(nullptr), m_objListPosX(-1), m_objListPosY(-1), m_objListWidth(-1), m_objListHeight(-1)
{
	setlocale(LC_CTYPE, "");
	wchar_t buffer[256] = L"MorphChecker";
	SetTitle(buffer);
	this->SetCloseButton(false);

	MQFrame *mainFrame = CreateVerticalFrame(this);
	mainFrame->SetVertLayout(LAYOUT_FILL);

	m_scrollbox = CreateScrollBox(mainFrame);
	m_scrollbox->SetVertLayout(LAYOUT_FILL);
	m_scrollbox->SetHorzLayout(LAYOUT_FILL);
	m_scrollbox->SetVertBarStatus(MQScrollBox::SCROLLBAR_ON);
	m_scrollbox->SetHorzBarStatus(MQScrollBox::SCROLLBAR_OFF);
	m_listframe = CreateVerticalFrame(m_scrollbox);
	m_listframe->SetHorzLayout(LAYOUT_FILL);
	m_listframe->SetOutSpace(0.0);

	{
		MQFrame *frame1 = CreateHorizontalFrame(mainFrame);
		m_button0 = CreateButton(frame1, L"0");
		m_button1 = CreateButton(frame1, L"50");
		m_button2 = CreateButton(frame1, L"100");
		m_button0->AddClickEvent(this, &MorphCheckerWindow::morph_change);
		m_button1->AddClickEvent(this, &MorphCheckerWindow::morph_change);
		m_button2->AddClickEvent(this, &MorphCheckerWindow::morph_change);
		m_morphvalue = CreateDoubleSpinBox(frame1);
		m_morphvalue->SetMin(0.0);
		m_morphvalue->SetMax(100.0);
		m_morphvalue->SetDecimalDigit(1);
		m_morphvalue->SetVisibleColumn(5);
		m_morphvalue->SetIncrement(1.0);
		m_morphvalue->AddChangingEvent(this, &MorphCheckerWindow::morph_change);
	}
	{
		MQFrame *frame2 = CreateHorizontalFrame(mainFrame);
		MQButton *button_clear = CreateButton(frame2, L"Clear");
		button_clear->AddClickEvent(this, &MorphCheckerWindow::clear);
		MQButton *button_objlist = CreateButton(frame2, L"Visible");
		button_objlist->AddClickEvent(this, &MorphCheckerWindow::toggleChildWindow);
		m_check_individual = CreateCheckBox(frame2, L"Objects");
		m_check_individual->AddChangedEvent(this, &MorphCheckerWindow::toggleIndividual);
		m_check_edit = CreateCheckBox(frame2, L"Edit");
		m_check_edit->AddChangedEvent(this, &MorphCheckerWindow::toggleEdit);
		m_check_edit->SetEnabled(false);
	}
	morphlistup();

	this->AddKeyUpEvent(this, &MorphCheckerWindow::keyevent);
	this->AddHideEvent(this, &MorphCheckerWindow::hideevent);
}

MorphCheckerWindow::~MorphCheckerWindow()
{
	delete m_scrollbox;
	m_scrollbox = nullptr;
	delete m_WindowObjlist;
	m_WindowObjlist = nullptr;
	clearScrollBox();
	m_morphmap.clear();
}

MQScrollBox *MorphCheckerWindow::CreateScrollBox(MQWidgetBase *parent)
{
	MQScrollBox *p = new MQScrollBox();
	p->SetAutoWidgetScroll(true);
	if(parent != NULL){
		parent->AddChild(p);
	}
	return p;
}

void MorphCheckerWindow::clearScrollBox()
{
	for(auto &item: m_morphlist) {
		m_listframe->RemoveChild(item.frame());
		delete item.frame();
	}
	m_morphlist.clear();
	m_current = nullptr;
}

MorphCheckerWindow::ListItem MorphCheckerWindow::addItemtoScrollBox(MQWidgetBase *parent, UINT uid, const std::wstring& name)
{
	MQFrame *frame = CreateHorizontalFrame(parent);
	frame->SetHorzLayout(LAYOUT_FILL);
	MQLabel *label = CreateLabel(frame, name);
	label->SetHorzLayout(LAYOUT_FILL);
	label->SetFontColor(192, 192, 192, 255);
	label->AddLeftDownEvent(this, &MorphCheckerWindow::morph_click);
	MQSlider *slider = CreateSlider(frame);
	slider->SetPosition(0.0);
	slider->SetMin(0.0);
	slider->SetMax(100.0);
	slider->AddChangingEvent(this, &MorphCheckerWindow::morph_change);
	ListItem item(uid, label, slider, frame);
	return item;
}

UINT MorphCheckerWindow::getCurrentMorph()
{
	return m_current->uid();
}

void MorphCheckerWindow::setMorphValue()
{
	if(m_current) {
		for(auto &morphobj : m_pPlugin->m_morphobjlist) {
			if(morphobj.getUniqueID() == m_current->uid()) {
				double value = morphobj.value();
				m_morphvalue->SetPosition(value);
				break;
			}
		}
	}
	else {
		m_morphvalue->SetPosition(0.0);
	}
}

BOOL MorphCheckerWindow::morphlistup()
{
	std::wstring oldname = m_current ? m_current->label()->GetText() : L"";
	m_morphmap.clear();
	clearScrollBox();
	bool individulal_mode = m_check_individual->GetChecked();
	int index = 0;
	for(auto &morphobj : m_pPlugin->m_morphobjlist) {
		// individulal_modeȂS[to^
		// łȂ΁A[tƂɓo^
		std::wstring morphname(!!m_check_individual->GetChecked() ? cmq::towstring(morphobj.getName().substr(5)) : morphobj.getMorphName());
		if(individulal_mode || m_morphmap.find(morphobj.getMorphName()) == m_morphmap.end()) {
			auto item = addItemtoScrollBox(m_listframe, morphobj.getUniqueID(), morphname);
			m_morphlist.push_back(item);
			double value = morphobj.value();
			item.slider()->SetPosition(value);
			if(!m_current && oldname == morphname) {
				m_current = &m_morphlist.back();
			}
		}
		m_morphmap.insert(std::make_pair(morphname, morphobj.getUniqueID()));
	}
	if(m_current) {
		m_current->select(true);
		m_current->setcurrent(true);
		setMorphValue();
	}
	return FALSE;
}

BOOL MorphCheckerWindow::listupdate(MQDocument doc)
{
	for(auto &item: m_morphlist) {
		if(item.isselected()) {
			std::wstring morphname = item.label()->GetText();
			double value = m_pPlugin->get_morph_value(item.uid());
			item.slider()->SetPosition(value);
			m_pPlugin->setSelectingMorph(item.uid());
		}
	}
	if(m_WindowObjlist) {
		m_WindowObjlist->objectlistup(nullptr);
	}
	return TRUE;
}

BOOL MorphCheckerWindow::toggleChildWindow(MQWidgetBase *sender, MQDocument doc)
{
	if (m_WindowObjlist == nullptr || !m_WindowObjlist->GetVisible()){
		openChild();
	}
	else {
		closeChild();
	}
	return FALSE;
}

BOOL MorphCheckerWindow::toggleIndividual(MQWidgetBase *sender, MQDocument doc)
{
	morphlistup();
	enableEditMode();
	return FALSE;
}

BOOL MorphCheckerWindow::toggleEdit(MQWidgetBase *sender, MQDocument doc)
{
	enableEditMode();
	return FALSE;
}

void MorphCheckerWindow::enableEditMode()
{
	if(!m_check_individual->GetChecked() || !m_current || m_current->slider()->GetPosition() == 0.0) {
		m_check_edit->SetEnabled(false);
		m_pPlugin->set_editmode(false);
	}
	else {
		m_check_edit->SetEnabled(true);
		m_pPlugin->set_editmode(m_check_edit->GetChecked());
		if(m_check_edit->GetChecked()) {
			for(auto &item : m_morphlist) {
				if(item.iscurrent()) {
					std::wstring morphname = item.label()->GetText();
					auto findit = m_morphmap.equal_range(morphname);
					for(auto it = findit.first; it != findit.second ; ++it) {
						UINT uid = it->second;
						m_pPlugin->setSelectingMorph(uid);
						break;
					}
				}
			}
		}
	}
}

BOOL MorphCheckerWindow::morph_click(MQWidgetBase *sender, MQDocument doc, MQWidgetMouseParam& param)
{
	if(MQLabel *click_label = dynamic_cast<MQLabel*>(sender)) {
		if(m_current && m_current->label() != click_label) {
			m_current->setcurrent(false);
		}
		auto findit = std::find_if(m_morphlist.begin(), m_morphlist.end(), [&click_label](ListItem &item) {return item.label() == click_label;});
		if(findit != m_morphlist.end()) {
			m_current = &(*findit);
			m_current->setcurrent(true);
			m_current->select(true);
			setMorphValue();
		}
		if(!param.Shift) {
			// ̑INA
			for(auto &item : m_morphlist) {
				if(m_current != &item) {
					item.select(false);
				}
			}
		}
	}
	enableEditMode();
	return FALSE;
}

BOOL MorphCheckerWindow::morph_change(MQWidgetBase *sender, MQDocument doc)
{
	double value = 0.0;
	MQSlider *click_slider = dynamic_cast<MQSlider*>(sender);
	MQDoubleSpinBox *spinbox = dynamic_cast<MQDoubleSpinBox*>(sender);
	if(click_slider) {
		value = click_slider->GetPosition();
		auto findit = std::find_if(m_morphlist.begin(), m_morphlist.end(), [&click_slider](ListItem &item) {return item.slider() == click_slider;});
		if(findit != m_morphlist.end()) {
			if(m_current) {
				m_current->setcurrent(false);
			}
			m_current = &(*findit);
			m_current->setcurrent(true);
			if(!m_current->isselected()) {
				// XC_[NbNACeIȂÃACe͔Iɂ
				for(auto &item : m_morphlist) {
					if(&item != m_current) {
						item.select(false);
					}
				}
			}
			m_current->select(true);
		}
	}
	else if(spinbox) {
		value = spinbox->GetPosition();
	}
	else {
		if(sender == m_button0) {
			value = 0.0;
		}
		else if(sender == m_button1) {
			value = 50.0;
		}
		else if(sender == m_button2) {
			value = 100.0;
		}
	}
	for(auto &item : m_morphlist) {
		if(item.isselected()) {
			std::wstring morphname = item.label()->GetText();
			auto findit = m_morphmap.equal_range(morphname);
			for(auto it = findit.first; it != findit.second ; ++it) {
				UINT uid = it->second;
				m_pPlugin->set_morph(uid, value);
			}
			MQSlider *slider = item.slider();
			slider->SetPosition(value);
			if(m_current == &item) {
				setMorphValue();
			}
		}
	}
	enableEditMode();
	return TRUE;
}

BOOL MorphCheckerWindow::clear(MQWidgetBase *sender, MQDocument doc)
{
	for(auto &morphobj : m_pPlugin->m_morphobjlist) {
		morphobj.setvalue(0.0);
	}
	for(auto &item : m_morphlist) {
		item.slider()->SetPosition(0.0);
		item.select(false);
		if(m_current == &item) {
			setMorphValue();
		}
	}
	m_pPlugin->m_morphupdate = true;
	enableEditMode();
	MQ_RefreshView(nullptr);
	return TRUE;
}

BOOL MorphCheckerWindow::keyevent(MQWidgetBase *sender, MQDocument doc, MQWidgetKeyParam& param)
{
	switch(param.Key) {
	case 0x43: clear(this, doc); break;
	case 0x56: toggleChildWindow(this, doc); break;
	case 0x31: return morph_change(m_button0, doc);
	case 0x32: return morph_change(m_button1, doc);
	case 0x33: return morph_change(m_button2, doc);
	}
	return FALSE;
}

BOOL MorphCheckerWindow::hideevent(MQWidgetBase *sender, MQDocument doc)
{
	closeChild();
	return FALSE;
}

void MorphCheckerWindow::openChild()
{
	if(!m_WindowObjlist) {
		m_WindowObjlist = new MorphCheckerObjectListWindow(m_pPlugin, *this);
	}
	if(m_objListPosX >= 0 && m_objListPosY >= 0) {
		m_WindowObjlist->SetPosX(m_objListPosX);
		m_WindowObjlist->SetPosY(m_objListPosY);
		m_WindowObjlist->SetWidth(m_objListWidth);
		m_WindowObjlist->SetHeight(m_objListHeight);
	}
	else {
		m_WindowObjlist->SetPosX(this->GetPosX() + this->GetWidth());
		m_WindowObjlist->SetPosY(this->GetPosY() + 20);
	}
	m_WindowObjlist->SetVisible(true);
}

void MorphCheckerWindow::closeChild()
{
	if(m_WindowObjlist) {
		m_objListPosX = m_WindowObjlist->GetPosX();
		m_objListPosY = m_WindowObjlist->GetPosY();
		m_objListWidth = m_WindowObjlist->GetWidth();
		m_objListHeight = m_WindowObjlist->GetHeight();
		m_WindowObjlist->SetVisible(false);
	}
}

//---------------------------------------------------------------------------------------------------

MorphCheckerObjectListWindow::MorphCheckerObjectListWindow(MorphChecker *plugin, MQWindowBase& parent)
	: MQWindow(parent), m_pPlugin(plugin), m_pParent(dynamic_cast<MorphCheckerWindow*>(&parent))
{
	setlocale(LC_CTYPE, "");
	wchar_t buffer[256] = L"MorphChecker";
	SetTitle(buffer);
//	this->SetCloseButton(false);

	MQFrame *mainFrame = CreateVerticalFrame(this);
	mainFrame->SetVertLayout(LAYOUT_FILL);

	m_list = CreateListBox(mainFrame);
	m_list->SetVertLayout(LAYOUT_FILL);
//	m_list->SetMultiSelect(false);
	m_list->AddChangedEvent(this, &MorphCheckerObjectListWindow::listchange);

	MQFrame *frame2 = CreateHorizontalFrame(mainFrame);
	MQButton *button_all = CreateButton(frame2, L"All");
	button_all->AddClickEvent(this, &MorphCheckerObjectListWindow::sel_all);
	MQButton *button_invert = CreateButton(frame2, L"Invert");
	button_invert->AddClickEvent(this, &MorphCheckerObjectListWindow::sel_inv);

	objectlistup(m_list);

	this->AddKeyUpEvent(this, &MorphCheckerObjectListWindow::keyevent);
	this->AddHideEvent(this, &MorphCheckerObjectListWindow::hideevent);
}

MorphCheckerObjectListWindow::~MorphCheckerObjectListWindow()
{
}

BOOL MorphCheckerObjectListWindow::setvisible(int index, bool visible)
{
	std::wstring name = m_list->GetItem(index);
	name[0] = visible ? '>' : ' ';
	m_list->SetItem(index, name);
	return FALSE;
}


BOOL MorphCheckerObjectListWindow::objectlistup(MQWidgetBase *parent) {
	int index = 0;
	for(auto &baseobj : m_pPlugin->m_baseobjlist) {
		std::wstring objname(L">");
		objname +=  baseobj.getObjName();
		m_list->AddItem(objname, baseobj.getUniqueID());
		setvisible(index++, baseobj.visible());
	}
	return FALSE;
}

BOOL MorphCheckerObjectListWindow::listchange(MQWidgetBase *parent, MQDocument doc)
{
	int index = m_list->GetCurrentIndex();
	if(m_list->GetItemSelected(index)) {
		std::wstring name = m_list->GetItem(index);
		UINT uid = (UINT)m_list->GetItemTag(index);
		bool visible = !m_pPlugin->get_visible(uid);
		m_pPlugin->set_visible(uid, visible);
		setvisible(index, visible);
		MQ_RefreshView(nullptr);
	}
	return FALSE;
}

BOOL MorphCheckerObjectListWindow::sel_all(MQWidgetBase *sender, MQDocument doc)
{
	for(int index = 0; index < m_list->GetItemCount(); index++) {
		UINT uid = (UINT)m_list->GetItemTag(index);
		m_pPlugin->set_visible(uid, true);
		setvisible(index, true);
	}
	MQ_RefreshView(nullptr);
	return FALSE;
}

BOOL MorphCheckerObjectListWindow::sel_inv(MQWidgetBase *sender, MQDocument doc)
{
	for(int index = 0; index < m_list->GetItemCount(); index++) {
		UINT uid = (UINT)m_list->GetItemTag(index);
		bool visible = !m_pPlugin->get_visible(uid);
		m_pPlugin->set_visible(uid, visible);
		setvisible(index, visible);
	}
	MQ_RefreshView(nullptr);
	return FALSE;
}

BOOL MorphCheckerObjectListWindow::keyevent(MQWidgetBase *sender, MQDocument doc, MQWidgetKeyParam& param)
{
	switch(param.Key) {
	case 0x1B: this->SetVisible(false);
	case 0x41: return this->sel_all(sender, doc);
	case 0x58: return this->sel_inv(sender, doc);
	}
	return FALSE;
}

BOOL MorphCheckerObjectListWindow::hideevent(MQWidgetBase *sender, MQDocument doc)
{
	m_pParent->closeChild();
	return FALSE;
}
