#include <map>
#include <unordered_map>

#include "MQWidget.h"
#include "CMQObject.h"

#ifndef _UIWindowDefine
#define _UIWindowDefine

class MorphChecker;
class MorphCheckerObjectListWindow;

class MorphCheckerWindow : public MQWindow
{
	friend class MorphCheckerObjectListWindow;
public:
	MorphCheckerWindow(MorphChecker *plugin, MQWindowBase& parent);
	virtual ~MorphCheckerWindow();

	struct ListItem
	{
	public:
		ListItem(UINT uid, MQLabel *label, MQSlider *slider, MQFrame *frame) : m_uid(uid), m_label(label), m_slider(slider), m_frame(frame), m_selected(false) {}
		~ListItem() {}
		void select(bool flag) {
			m_selected = flag;
			m_label->SetFontBold(flag);
			if(flag) {
				m_label->SetFontColor(255, 255, 0, 255);
			}
			else {
				m_label->SetFontColor(192, 192, 192, 255);
			}
			m_label->Repaint(true);
		}
		bool isselected() const {return m_selected;}
		void setcurrent(bool flag) {m_label->SetFrame(flag); m_label->Repaint();}
		bool iscurrent() const {return m_label->GetFrame();}
		UINT uid() {return m_uid;}
		MQLabel *label() {return m_label;}
		MQSlider *slider() {return m_slider;}
		MQFrame *frame() {return m_frame;}
	private:
		UINT m_uid;
		MQLabel *m_label;
		MQSlider *m_slider;
		MQFrame *m_frame;
		bool m_selected;
	};

	BOOL morphlistup();
	BOOL listupdate(MQDocument doc);
	BOOL set_status(bool busy);

	UINT getCurrentMorph();

private:
	MorphChecker *m_pPlugin;
	MorphCheckerObjectListWindow *m_WindowObjlist;

	ListItem *m_current;
	int m_objListPosX, m_objListPosY, m_objListWidth, m_objListHeight;
	std::unordered_multimap<std::wstring, UINT> m_morphmap;
	std::vector<ListItem> m_morphlist;

//	MQListBox *m_list;
	MQScrollBox *m_scrollbox;
	MQFrame *m_listframe;
//	MQSlider *m_slider;
	MQButton *m_button0;
	MQButton *m_button1;
	MQButton *m_button2;
	MQDoubleSpinBox *m_morphvalue;
	MQCheckBox *m_check_individual;
	MQCheckBox *m_check_edit;

	void openChild();
	void closeChild();

	void setMorphValue();
//	BOOL setvalue(int index, double value);
	BOOL clear(MQWidgetBase *sender, MQDocument doc);
	BOOL listchange(MQWidgetBase *sender, MQDocument doc);
	BOOL toggleChildWindow(MQWidgetBase *sender, MQDocument doc);
	BOOL toggleIndividual(MQWidgetBase *sender, MQDocument doc);
	BOOL toggleEdit(MQWidgetBase *sender, MQDocument doc);
	void enableEditMode();

	BOOL morph_click(MQWidgetBase *sender, MQDocument doc, MQWidgetMouseParam& param);
	BOOL morph_change(MQWidgetBase *sender, MQDocument doc);
	BOOL keyevent(MQWidgetBase *sender, MQDocument doc, MQWidgetKeyParam& param);
	BOOL hideevent(MQWidgetBase *sender, MQDocument doc);

	MQScrollBox *CreateScrollBox(MQWidgetBase *parent);
	void clearScrollBox();
	ListItem addItemtoScrollBox(MQWidgetBase *parent, UINT uid, const std::wstring& name);
};

class MorphCheckerObjectListWindow : public MQWindow
{
public:
	MorphCheckerObjectListWindow(MorphChecker *plugin, MQWindowBase &parent);
	~MorphCheckerObjectListWindow();

	BOOL objectlistup(MQWidgetBase *parent);
private:
	MorphChecker *m_pPlugin;
	MorphCheckerWindow *m_pParent;

	MQListBox *m_list;

	MQScrollBox *CreateScrollBox(MQWidgetBase *parent);

	BOOL setvisible(int index, bool visible);

	BOOL listchange(MQWidgetBase *parent, MQDocument doc);
	BOOL sel_all(MQWidgetBase *sender, MQDocument doc);
	BOOL sel_inv(MQWidgetBase *sender, MQDocument doc);
	BOOL keyevent(MQWidgetBase *sender, MQDocument doc, MQWidgetKeyParam& param);
	BOOL hideevent(MQWidgetBase *sender, MQDocument doc);
};

#endif
